/*
Authors: Emiliano Cristiani and Marta Menci, September 2023
RELATED PAPER: https://doi.org/10.1016/j.physa.2023.128625
Please acknowledge the authors of the paper (E. Cristiani, M. Menci, A. Malagnino, G.G. Amaro) and GAe Engineering s.r.l. for funding the research
*/

/*
Parameters are in files 'parameters.dat' and 'geometry.dat'
Parameters explanations are in files 'parameters.info' and 'geometry.info'
Units are 'meter' and 'second'
*/

/*
This code reproduces the Test 1a in the paper (corridor with closed gate at the right side). Pushing distance and physical distance are slightly lower here.
If you want to use exactly the same parameters as in the paper you have to prevent that "P[k].ispushed" becomes true at any time (line 178).
*/

#include<cstdlib>
#include<fstream>
#include<time.h>
#include<math.h>
#include<iostream>
using namespace std;

///define
#define pi 3.14159
#define RANDOMVAL   ( ((rand()%10000)+1.0) / 10000.0 )  //uniform random variable in (0,1]

///structs
struct vect{
 double x;
 double y;
};

struct PEDESTRIAN{
 vect   pos;
 bool   iscontacted;
 bool   ispushed;
 double desireddist;
};

///prototypes
vect MAKE_CHOICE(int);
vect BE_PUSHED(int,int,double);
vect TRY_SLIDING(int,double,double);
bool IS_IN_THE_DOMAIN(vect);
vect FIND_ADMISSIBLE_INITIAL_CONDITION(int,double);
void PLOT_RUNTIME(FILE*,int);

/// global variables
PEDESTRIAN *P;       //vector containing all the pedestrians
PEDESTRIAN **SYSTEM; //matrix containing the complete history of the simulation
int Np;              //number of pedestrians
double L,Lreduced,H; //corridor size
double Dt;           //time step
double Dx;           //space step


/// ////////////////////////////  MAIN  /////////////////////////////

int main(){

printf("Inizialization...\n");

///local variables (parameters)
int    Nt;                         //numbers of time steps
double Tfinal;                     //final time
bool   plot_gnuplot=false;         //use this if you have GNUplot installed and you want to see the simulation runtime
int    DnPlot=10;                  //GNUplot is called every DnPlot time steps
bool   save_on_file=true;          //save results on file (to plot them afterwards)
double velcomfort,C,alpha,epsilon;
double comfortdist,contactdist,pushingdist,physicaldist; //inter-personal distances (parameters)

///other local variables
FILE *pf,*pgnuplot;                           //pointers to files
int n;                                        //time counter
int k,kk,kkback,kkfront;                      //indices for pedestrians
double dist,mindist,mindistback,mindistfront; //inter-personal distances

///read parameters from file 'parameters.dat'
pf=fopen("parameters.dat","r"); if (pf==NULL) {printf("file parameters.dat non found\n"); return 0;}
fscanf(pf,"%i\n", &Np);
fscanf(pf,"%lf\n",&Dt);
fscanf(pf,"%lf\n",&Tfinal);
fscanf(pf,"%lf\n",&velcomfort);
fscanf(pf,"%lf\n",&comfortdist);
fscanf(pf,"%lf\n",&contactdist);
fscanf(pf,"%lf\n",&pushingdist);
fscanf(pf,"%lf\n",&physicaldist);
fscanf(pf,"%lf\n",&alpha);
fscanf(pf,"%lf\n",&epsilon);
fscanf(pf,"%lf\n",&C);
fclose(pf);

///read parameters from file 'geometry.dat'
pf=fopen("geometry.dat","r"); if (pf==NULL) {printf("file geometry.dat non found\n"); return 0;}
fscanf(pf,"%lf\n",&L);
fscanf(pf,"%lf\n",&Lreduced);
fscanf(pf,"%lf\n",&H);
fclose(pf);

///derived parameters
Nt=(int)round(Tfinal/Dt); //number of time steps
Dx=velcomfort*Dt;         //space step

///check & write parameters
printf("Number of pedestrians Np=%i\n",Np);
printf("Time step Dt=%1.2lf sec\n",Dt);
printf("Final time=%1.1lf sec\n",Tfinal);
printf("comfort velocity=%1.2lf m/s\n",velcomfort);
printf("comfort distance=%1.2lf m\n",comfortdist);
printf("contact distance=%1.2lf m\n",contactdist);
printf("pushing distance=%1.2lf m\n",pushingdist);
printf("physical distance=%1.2lf m\n",physicaldist);
printf("alpha=%1.2lf\n",alpha);
printf("epsilon=%1.2lf\n",epsilon);
printf("C=%1.2lf\n",C);
printf("Number of time steps Nt=%i\n",Nt);
printf("CORRIDOR: L=%1.1lf, Lreduced=%1.1lf, H=%1.1lf\n",L,Lreduced,H);

///set seed random number generator
//srand(time(NULL)); //random not repeatable
srand(1); //random repeatable

///initialization of SYSTEM and P
SYSTEM=(PEDESTRIAN**)malloc(Nt*sizeof(PEDESTRIAN*));
for(n=0;n<Nt;n++) {SYSTEM[n]=(PEDESTRIAN*)malloc(Np*sizeof(PEDESTRIAN));}

P=(PEDESTRIAN*)malloc(Np*sizeof(PEDESTRIAN));
printf("Finding initial position of pedestrians... \n(if the code frezees here probably there is no enough space for pedestrians in the domain)\n");
for(k=0;k<Np;k++){
  P[k].pos=FIND_ADMISSIBLE_INITIAL_CONDITION(k,physicaldist); //here we set the initial positions of pedestrians
  P[k].desireddist=comfortdist;
  P[k].iscontacted=false;
  P[k].ispushed=false;
}
printf("Done\n");

///inizialize GNUplot
if (plot_gnuplot){
  pgnuplot=popen("gnuplot -persist","w");
  fprintf(pgnuplot,"set xrange[0:%lf]; set yrange[0:%lf]; set size ratio %lf; set size 1; set terminal wxt size 1700,400; unset key\n",L,H,H/L); //for windows
  //fprintf(pgnuplot,"set xrange[0:%lf]; set yrange[0:%lf]; set size ratio %lf; set size 1; set terminal qt size 1700,400; unset key\n",L,H,H/L);    //for linux
}

printf("\nStart computation...\n");

/// ///////////  BEGIN CYCLE IN TIME //////////////
for(n=0;n<Nt;n++){

  printf("time step n=%i/%i\n",n+1,Nt);

  ///photo of the system's status at time n
  for(k=0;k<Np;k++){
    SYSTEM[n][k].pos.x       =P[k].pos.x;
    SYSTEM[n][k].pos.y       =P[k].pos.y;
    SYSTEM[n][k].iscontacted =P[k].iscontacted;
    SYSTEM[n][k].ispushed    =P[k].ispushed;
    SYSTEM[n][k].desireddist =P[k].desireddist;
  }

  ///MAIN CYCLE ON PEDESTRIANS
  for(k=0;k<Np;k++){

    ///compute distances from neighbors (front, back and min among the two)
    mindistback=1e10; mindistfront=1e10;
    kkback=-1;        kkfront=-1;
    for(kk=0;kk<Np;kk++){
      if (kk!=k){
        dist=sqrt( (P[k].pos.x-P[kk].pos.x)*(P[k].pos.x-P[kk].pos.x) + (P[k].pos.y-P[kk].pos.y)*(P[k].pos.y-P[kk].pos.y) );
        if (P[k].pos.x<=P[kk].pos.x){ //kk is in front of k
          if (dist<mindistfront) {mindistfront=dist; kkfront=kk;} }
        else{                         //kk is behind k
          if (dist<mindistback)  {mindistback=dist;  kkback=kk;}
        }
      }
    }
    if (mindistback<mindistfront) mindist=mindistback; else mindist=mindistfront;
    if (mindist==1e10) {printf("ERROR: mindist=1e10\n"); exit(0);} //check DEBUG

    ///check contact
    if (mindist<contactdist) P[k].iscontacted=true;  else P[k].iscontacted=false;

    ///check pushing
    if (mindistback<pushingdist) P[k].ispushed=true; else P[k].ispushed=false;

    ///update desireddist with bounds
    if (mindistback<alpha*mindistfront){   //magic condition
      P[k].desireddist=mindistback;
      if (P[k].desireddist>comfortdist) P[k].desireddist=comfortdist; //upper bound
      if (P[k].desireddist<contactdist) P[k].desireddist=contactdist; //lower bound
    }

    ///update pedestrian positions
    if (!P[k].ispushed){

      P[k].pos=MAKE_CHOICE(k);} //proceed normally with BLOCK 1 (see related paper)

    else{ //is pushed

      if (mindistfront>physicaldist){
        P[k].pos=BE_PUSHED(k,kkback,C);}          //BLOCK 2 (see related paper)
      else{
        P[k].pos=TRY_SLIDING(k,mindist,epsilon);  //BLOCK 3 (see related paper)
      }
    }

    ///horizontal obstacles
    if (P[k].pos.y>=H-0.3)   P[k].pos.y=H-0.3;   //upper
    if (P[k].pos.y<=0.0+0.3) P[k].pos.y=0.0+0.3; //lower

  } //end for k

  //plot runtime
  if ((n%DnPlot==0) && (plot_gnuplot)) PLOT_RUNTIME(pgnuplot,n);

} //end for n

/// ///////////  END CYCLE IN TIME //////////////

printf("\nEnd of computation\n");

///final check (consistency)
{
 int people_found=0;
 for(k=0;k<Np;k++){
   if (IS_IN_THE_DOMAIN(P[k].pos)){
     people_found++;}
   else{
    printf("WARNING: AGENT OUT k=%i, x=%lf y=%lf\n",k,P[k].pos.x,P[k].pos.y);
   }
 }
 printf("CHECK: people inside the domain at final time=%i/%i\n",people_found,Np);
}


///save SYSTEM on file [time step | agent idx | pos_x | pos_y | iscontacted | ispushed | desireddist]
if (save_on_file){
printf("saving result...\n");
pf=fopen("result.out","w");
for(n=0;n<Nt;n++){
  for(k=0;k<Np;k++){
    fprintf(pf,"%i %i %1.2lf %1.2lf %i %i %1.4lf\n",
            n,                          //col 1
            k,                          //col 2
            SYSTEM[n][k].pos.x,         //col 3
            SYSTEM[n][k].pos.y,         //col 4
            SYSTEM[n][k].iscontacted,   //col 5
            SYSTEM[n][k].ispushed,      //col 6
            SYSTEM[n][k].desireddist);  //col 7
  }
}
fclose(pf);
printf("Result saved in 'result.out'\n");
}


///final stuff
if (plot_gnuplot) pclose(pgnuplot);
for(n=0;n<Nt;n++) free(SYSTEM[n]);
free(SYSTEM);
free(P);
printf("END\n");
return 1;
}


///
///
///
///
/// ////////////////// FUNCTIONS ///////////////////
///
///
///
///


//////////////////// BLOCK 1 ///////////////////////
vect MAKE_CHOICE(int k){
 int i,j,idx,app;
 vect newpos;
 double theta;
 double dist,dist_to_target_temp,min_dist_to_target,mindistfronttemp;
 vect temp;
 int Nc=17;          //number of possible new positions for agent k
 double Dtheta=10.0;
 bool moved=false;
 int V[Nc];

 ///random permutation of [0:Nc-1] numbers (Fisher-Yates algorithm)
 for (i=0;i<Nc;i++) {V[i]=i;}
 for (i=Nc-1;i>0;i--){
   j=rand()%(i+1);
   app=V[j]; V[j]=V[i]; V[i]=app;
 }

 ///starting tentatives
 min_dist_to_target=1e10;
 for (i=0;i<Nc;i++){

   idx=V[i];

   ///new tentative position (just right-half circle since it can only move forward towards the target)
   theta=(-80.0+idx*Dtheta)*pi/180.0; //degrees converted in radiants (-80:10:80)
   temp.x = P[k].pos.x + Dx*cos(theta);
   temp.y = P[k].pos.y + Dx*sin(theta);

    ///find mindistfronttemp (i.e. the distance from nearest neighbor ahead, from the tentative position)
    mindistfronttemp=1e10;
    for(int kk=0;kk<Np;kk++){
      if (kk!=k){
      	//if (temp.x<P[kk].pos.x){         //check if kk is in front of k, assuming k is in temp
        if (P[k].pos.x<P[kk].pos.x){       //check if kk is in front of k, assuming k is in P[k].pos (preferred choice)
          dist=sqrt( (temp.x-P[kk].pos.x)*(temp.x-P[kk].pos.x) + (temp.y-P[kk].pos.y)*(temp.y-P[kk].pos.y) );
          if (dist<mindistfronttemp){
            mindistfronttemp=dist;
          }
        }
      }
    }

   ///find the closest position to the target among the acceptable ones, within boundaries
   dist_to_target_temp=L-temp.x;
   if ( (mindistfronttemp>P[k].desireddist) && (IS_IN_THE_DOMAIN(temp)) ){
     if (dist_to_target_temp<min_dist_to_target){
       min_dist_to_target=dist_to_target_temp;
       newpos=temp;
       moved=true;
     }
   }

 }
 if (!moved) newpos=P[k].pos; //if cannot move it stands still

 return newpos;
}


//////////////////// BLOCK 2 ///////////////////////
vect BE_PUSHED(int k, int kkback, double C){
 vect newpos,pushingdirection;

 pushingdirection.x = P[k].pos.x - P[kkback].pos.x;
 pushingdirection.y = P[k].pos.y - P[kkback].pos.y;

 newpos.x = P[k].pos.x + C*Dt*pushingdirection.x;
 newpos.y = P[k].pos.y + C*Dt*pushingdirection.y;

 ///check boundaries
 if (!IS_IN_THE_DOMAIN(newpos)) {newpos=P[k].pos;}

 return newpos;
}


//////////////////// BLOCK 3 ///////////////////////
vect TRY_SLIDING(int k, double mindist, double epsilon){
 vect temp,newpos;
 int i,j,idx,app;
 int Nc=36;
 int V[Nc];
 double theta;
 double Dtheta=10.0;
 double dist,mindisttemp;
 double max_mindist;

 ///permutation of 0:Nc-1 numbers
 for (i=0;i<Nc;i++) {V[i]=i;}
 for (i=Nc-1;i>0;i--){
   j=rand()%(i+1);
   app=V[j]; V[j]=V[i]; V[i]=app;
 }

 ///starting tentatives
 max_mindist=0.0;
 for (i=0;i<Nc;i++){

   idx=V[i];

   ///new tentative position
   theta=(idx*Dtheta)*pi/180.0;
   temp.x = P[k].pos.x + epsilon*Dx*cos(theta);
   temp.y = P[k].pos.y + epsilon*Dx*sin(theta);

   ///find mindisttemp (i.e. the distance from nearest neighbor, from the tentative position)
   mindisttemp=1e10;
   for(int kk=0;kk<Np;kk++){
     if (kk!=k){
       dist=sqrt( (temp.x-P[kk].pos.x)*(temp.x-P[kk].pos.x) + (temp.y-P[kk].pos.y)*(temp.y-P[kk].pos.y) );
       if (dist<mindisttemp){
         mindisttemp=dist;
       }
     }
   }

   ///maximization and boundaries
   if ( (mindisttemp>max_mindist) && (IS_IN_THE_DOMAIN(temp)) ){
     max_mindist=mindisttemp;
     newpos=temp;
   }

 }

 ///maybe staying still is the better think to do?
 if (max_mindist<mindist){
  newpos=P[k].pos;
 }

 return newpos;
}


///////////////////////////////////////////////////////////////
void PLOT_RUNTIME(FILE *pgnuplot, int n){        //thanks to Simone Cacace, Sapienza Universita di Roma, for this function
 int pointsize,linecolor;
 fprintf(pgnuplot,"\nset xlabel 'time= %1.1lf sec'\n",n*Dt);

 //fprintf(pgnuplot,"plot '-' w l lw 2 lc rgb 'black', '-' using 1:2:3:4:5 w p pt var ps var lc rgb var\n");
 //fprintf(pgnuplot,"plot '-' w p pt 7 ps 2\n");
 fprintf(pgnuplot,"plot '-' using 1:2:3:4 w p pt 7 ps var lc var\n");
 for(int k=0;k<Np;k++){
   pointsize=1; linecolor=6;
   if (P[k].iscontacted) {pointsize=1; linecolor=4;}
   if (P[k].ispushed)    {pointsize=1; linecolor=7;}
   fprintf(pgnuplot,"%lf %lf %i %i\n",P[k].pos.x,P[k].pos.y,pointsize,linecolor);
 }
 fprintf(pgnuplot,"e\n");

 fflush(pgnuplot);
}


//////////////////////////////////////////////////////////////
vect FIND_ADMISSIBLE_INITIAL_CONDITION(int k, double thresholddist){
 double dist,mindist;
 vect POS;

 if (k==0){
  POS.x=RANDOMVAL*Lreduced+(L-Lreduced-2.0);
  POS.y=RANDOMVAL*(H-2.0)+1.0;
  return POS;
 }

 mindist=0.0;
 while (mindist<thresholddist){ //ATTENTION! This loop could be endless!

   POS.x=RANDOMVAL*Lreduced+(L-Lreduced-2);
   POS.y=RANDOMVAL*(H-2)+1;

   mindist=1e10;
   for(int kk=0;kk<k;kk++){ //all pedestrians before k, excluded k
    dist=sqrt( (POS.x-P[kk].pos.x)*(POS.x-P[kk].pos.x) + (POS.y-P[kk].pos.y)*(POS.y-P[kk].pos.y) );
     if (dist<mindist){
       mindist=dist;
     }
   }
 }
 return POS;
}


//////////////////////////////////////////////////////////////
bool IS_IN_THE_DOMAIN(vect POS){
 if ( (POS.x<L) && (POS.x>0.0) && (POS.y<H) && (POS.y>0.0) ){
   return true;
 }
 return false;
}
